<?php

namespace Tablesome\Includes\Modules\TablesomeDB_Rest_Api;

if (!defined('ABSPATH')) {
    exit; // Exit if accessed directly.
}

if (!class_exists('\Tablesome\Includes\Modules\TablesomeDB_Rest_Api\TablesomeDB_Rest_Api')) {
    class TablesomeDB_Rest_Api
    {
        public $tablesome_db;

        public function init()
        {
            $namespece = 'tablesome/v1';

            $this->tablesome_db = new \Tablesome\Includes\Modules\TablesomeDB\TablesomeDB();

            /** All REST-API Routes */
            $routes_controller = new \Tablesome\Includes\Modules\TablesomeDB_Rest_Api\Routes();
            $routes = $routes_controller->get_routes();

            foreach ($routes as $route) {
                /** Register the REST route */
                register_rest_route($namespece, $route['url'], $route['args']);
            }

        }

        public function api_access_permission()
        {
            if (current_user_can('edit_posts')) {
                return true;
            }
            $error_code = "UNAUTHORIZED";
            return new \WP_Error($error_code, $this->get_error_message($error_code));
        }

        public function get_error_message($error_code)
        {
            $messages = array(
                'UNAUTHORIZED' => "You don't have an permission to access this resource",
                'REQUIRED_POST_ID' => "Required, Tablesome table ID ",
                'INVALID_POST' => "Invalid, Tablesome post",
                'REQUIRED_RECORD_IDS' => "Required, Tablesome table record IDs",
                'UNABLE_TO_CREATE' => "Unable to create a post.",
            );

            $message = isset($messages[$error_code]) ? $messages[$error_code] : 'Something Went Wrong, try later';
            return $message;
        }

        public function create_or_update_table($request)
        {
            $params = $request->get_params();
            $table = new \Tablesome\Includes\Core\Table();
            $table_id = isset($params['table_id']) ? $params['table_id'] : 0;
            $columns = isset($params['columns']) ? $params['columns'] : [];

            $post_data = array(
                'post_title' => isset($params['table_title']) ? $params['table_title'] : 'Untitled Table',
                'post_type' => TABLESOME_CPT,
                'post_content' => isset($params['content']) ? $params['content'] : '',
                'post_status' => isset($params['table_status']) ? $params['table_status'] : 'publish',
            );
            $table_id = $table->insert_or_update_post($table_id, $post_data);
            if (empty($table_id)) {
                $response = array(
                    'status' => 'failed',
                    'message' => $this->get_error_message('UNABLE_TO_CREATE'),
                );
                return rest_ensure_response($response);
            }

            $table->set_table_meta_data($table_id,
                array(
                    'columns' => $columns,
                )
            );

            $meta_data = get_tablesome_data($table_id);

            $response = array(
                'table_id' => $table_id,
                'table_meta' => $meta_data,
                'status' => 'success',
            );
            return rest_ensure_response($response);
        }

        public function get_tables($request)
        {
            $data = array();
            /** Get all tablesome posts */
            $posts = get_posts(
                array(
                    'post_type' => TABLESOME_CPT,
                    'numberposts' => -1,
                )
            );
            $response_data = array(
                'data' => $data,
                'message' => 'Get all tablesome tables data',
            );

            if (empty($posts)) {
                return rest_ensure_response($response_data);
            }
            $tablesome_db = new \Tablesome\Includes\Modules\TablesomeDB\TablesomeDB();

            foreach ($posts as $post) {
                $meta_data = get_tablesome_data($post->ID);

                $table = $tablesome_db->create_table_instance($post->ID);
                /** Get records count */
                $records_count = $table->count();

                $data[] = array(
                    'ID' => $post->ID,
                    'post_title' => $post->post_title,
                    'post_content' => $post->post_title,
                    'post_status' => $post->post_status,
                    'meta_data' => $meta_data,
                    'records_count' => $records_count,
                );
            }

            $response_data['data'] = $data;
            return rest_ensure_response($data);
        }

        public function get_table_data($request)
        {
            $data = array();
            $table_id = $request->get_param('table_id');
            $post = get_post($table_id);

            if (empty($post) || $post->post_type != TABLESOME_CPT) {
                $error_code = "INVALID_POST";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }
            $tablesome_db = new \Tablesome\Includes\Modules\TablesomeDB\TablesomeDB();
            $meta_data = get_tablesome_data($post->ID);

            $table = $tablesome_db->create_table_instance($post->ID);
            $records_count = $table->count();

            $query = $tablesome_db->query(array(
                'table_id' => $post->ID,
                'table_name' => $table->name,
                'orderby' => 'rank_order',
                'order' => 'asc',
            ));

            $data = array(
                'ID' => $post->ID,
                'post_title' => $post->post_title,
                'post_content' => $post->post_content,
                'post_status' => $post->post_status,
                'meta_data' => $meta_data,
                'records' => $query->items,
                'records_count' => $records_count,
                'status' => 'success',
                'message' => 'Successfully get table with records',
            );

            return rest_ensure_response($data);
        }

        public function delete($request)
        {
            $table_id = $request->get_param('table_id');

            if (empty($table_id)) {
                $error_code = "REQUIRED_POST_ID";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }

            $post = get_post($table_id);

            if (empty($post) || $post->post_type != TABLESOME_CPT) {
                $error_code = "INVALID_POST";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }
            $table = $this->tablesome_db->create_table_instance($post->ID);
            $table_drop = $table->drop();

            $message = 'Table Deleted';
            if (!$table_drop) {
                $message = 'Can\'t delete the table';
            }

            $response_data = array(
                'message' => $message,
            );
            return rest_ensure_response($response_data);
        }

        public function get_table_records($request)
        {
            $params = $request->get_params();

            $table_id = isset($params['table_id']) ? $params['table_id'] : 0;

            if (empty($table_id)) {
                $error_code = "REQUIRED_POST_ID";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }

            $query_args = isset($params['query_args']) && is_array($params['query_args']) ? $params['query_args'] : [];

            $post = get_post($table_id);

            if (empty($post) || $post->post_type != TABLESOME_CPT) {
                $error_code = "INVALID_POST";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }

            $tablesome_db = new \Tablesome\Includes\Modules\TablesomeDB\TablesomeDB();
            $table = $tablesome_db->create_table_instance($post->ID);

            $query_args = array_merge(
                array(
                    'table_id' => $post->ID,
                    'table_name' => $table->name,
                )
                , $query_args);

            $query = $tablesome_db->query($query_args);

            // TODO: Return the formatted data if need. don't send the actual db data
            $records = $query->items;

            $response_data = array(
                'records' => $records,
                'message' => 'Get records successfully',
                'status' => 'success',
            );

            return rest_ensure_response($response_data);
        }

        public function modified_records($request)
        {
            $params = $request->get_params();

            $table_id = isset($params['table_id']) ? $params['table_id'] : 0;

            if (empty($table_id)) {
                $error_code = "REQUIRED_POST_ID";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }

            $post = get_post($table_id);

            if (empty($post) || $post->post_type != TABLESOME_CPT) {
                $error_code = "INVALID_POST";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }

            $records_inserted = isset($params['records_inserted']) ? json_decode($params['records_inserted'], true) : [];
            $records_updated = isset($params['records_updated']) ? json_decode($params['records_updated'], true) : [];

            /** Don't do anything if the inserts & updates records are empty  */
            $is_empty_records = (empty($records_inserted) && empty($records_updated));
            if ($is_empty_records) {
                return rest_ensure_response(array(
                    'status' => 'success',
                    'message' => 'Records Empty',
                ));
            }

            $tablesome_db = new \Tablesome\Includes\Modules\TablesomeDB\TablesomeDB();
            $table = $tablesome_db->create_table_instance($table_id);

            $query = $tablesome_db->query(array(
                'table_id' => $table_id,
                'table_name' => $table->name,
            ));

            /** Table MetaData */
            $meta_data = get_tablesome_data($table_id);

            $inserted_records_count = 0;
            $updated_records_count = 0;

            /** Insert all records  */
            if (!empty($records_inserted) && is_array($records_inserted)) {
                $insert_info = $tablesome_db->insert_many($table_id, $meta_data, $records_inserted);
                $inserted_records_count = isset($insert_info) && $insert_info['records_inserted_count'] ? $insert_info['records_inserted_count'] : 0;
            }

            // TODO: Need implement updating bulk record
            /**  */
            if (!empty($records_updated) && is_array($records_updated)) {

                foreach ($records_updated as $record) {

                    $record_id = isset($record['record_id']) ? $record['record_id'] : 0;

                    $data = $tablesome_db->table_crud_wp->helper->get_column_ided_record($meta_data, $record);
                    $data['post_id'] = $table_id;

                    // Insert the record if the record id value is 0
                    if (empty($record_id)) {
                        $insert_record = $tablesome_db->insert($query, $data);

                        if ($insert_record) {
                            $inserted_records_count = ++$inserted_records_count;
                        }
                    }

                    if (!empty($record_id) && intval($record_id)) {
                        $update_record = $tablesome_db->update_record($query, $record_id, $data);

                        if ($update_record) {
                            $updated_records_count = ++$updated_records_count;
                        }
                    }
                }
            }

            $response_data = array(
                'inserted_records_count' => $inserted_records_count,
                'updated_records_count' => $updated_records_count,
                'message' => 'Records modified successfully',
                'status' => 'success',
            );
            return rest_ensure_response($response_data);
        }

        public function delete_records($request)
        {
            $params = $request->get_params();
            $table_id = $request->get_param('table_id');
            if (empty($table_id)) {
                $error_code = "REQUIRED_POST_ID";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }

            $record_ids = $request->get_param("record_ids");

            $post = get_post($table_id);

            if (empty($post) || $post->post_type != TABLESOME_CPT) {
                $error_code = "INVALID_POST";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }

            if (empty($record_ids)) {
                $error_code = "REQUIRED_RECORD_IDS";
                return new \WP_Error($error_code, $this->get_error_message($error_code));
            }

            $message = 'Records removed successfully';

            $tablesome_db = new \Tablesome\Includes\Modules\TablesomeDB\TablesomeDB();
            $table = $tablesome_db->create_table_instance($post->ID);

            $query = $tablesome_db->query(array(
                'table_id' => $post->ID,
                'table_name' => $table->name,
            ));

            $delete_records = $tablesome_db->delete_records($query, $record_ids);

            $response_data = array(
                'message' => $message,
                'status' => ($delete_records) ? 'success' : 'failed',
            );
            return rest_ensure_response($response_data);
        }

    }
}